/*
 * The software provided here is released by the National
 * Institute of Standards and Technology (NIST), an agency of
 * the U.S. Department of Commerce, Gaithersburg MD 20899,
 * USA.  The software bears no warranty, either expressed or
 * implied. NIST does not assume legal liability nor
 * responsibility for a User's use of the software or the
 * results of such use.
 *
 * Please note that within the United States, copyright
 * protection, under Section 105 of the United States Code,
 * Title 17, is not available for any work of the United
 * States Government and/or for any works created by United
 * States Government employees. User acknowledges that this
 * software contains work which was created by NIST employees
 * and is therefore in the public domain and not subject to
 * copyright.  The User may use, distribute, or incorporate
 * this software provided the User acknowledges this via an
 * explicit acknowledgment of NIST-related contributions to
 * the User's work. User also agrees to acknowledge, via an
 * explicit acknowledgment, that any modifications or
 * alterations have been made to this software before
 * redistribution.
 * --------------------------------------------------------------------
 *
 * Revision History:
 * 2009 Jan Kyle Sanders & Simson Garfinkel: Extended
 * 2008 Nov Kyle Sanders - Created
 */


#include "ataraw.h"

#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>


int success = 0;
int fails = 0;

void print_test( int ret, char* test_name ){
    printf("%s test:  %s\n", test_name, (ret == ATA_RAW_OK)?"SUCCESSFUL":"FAILED");
    if(ret==ATA_RAW_OK){
	success++;
    }
    else {
	fails++;
    }
}

/* Try a read/write test where we read a sector, change it, write it back,
 * change it back, and verify each time.
 * This requires a drive that is 100MB in size
 */

int max=18;
int read_write_sector_test(int fd)
{
    u_char *orig_sector[max];
    u_char *new_sector[max];
    int i;
    u_long sector_number;
    int ret;
    int errors = 0;
    for(i=0;i<max;i++){
	orig_sector[i] = malloc(SECTOR_SIZE);
	new_sector[i] = calloc(SECTOR_SIZE,1);
	if(orig_sector[i]==NULL || new_sector[i]==NULL){
	    printf("malloc failed");
	    return ATA_RAW_NO_MEM;
	}
	sector_number = 1<<i;
	printf("reading sector %lu\n",sector_number);
	ret = ata_read_sector(fd,sector_number,1,orig_sector[i]);
	if (ret!=ATA_RAW_OK){
	    printf("cannot read sector %lu; error code=%d\n",sector_number,ret);
	    return ret; /* error reading? */
	}
	sprintf((char *)new_sector[i],"This is sector %lu\n",sector_number);
    }

    /* Now lets make the changes... ...  From this point on, we will need to restore
     * the sectors if an error is encountered...
     */
    for(i=0;i<max;i++){
	sector_number = 1<<i;
	printf("Writing sector %lu\n",sector_number);
	ret = ata_write_sector(fd,sector_number,1,new_sector[i]);
	if(ret != ATA_RAW_OK){
	    printf("error encountered writing sector %lu; error code=%d\n",sector_number,ret);
	    errors ++;
	}
    }

    /* Now verify the sectors. We have to write them all, and then verify.
     * Otherwise all read/write operaitons could be going to the same sector.
     */
    for(i=0;i<max;i++){
	u_char buf[SECTOR_SIZE];
	sector_number = 1<<i;
	ret = ata_read_sector(fd,sector_number,1,buf);
	if(ret != ATA_RAW_OK){
	    printf("error encountered reading sector %lu; error code=%d\n",sector_number,ret);
	    errors ++;
	} else {
	    if(memcmp(buf,new_sector[i],SECTOR_SIZE)!=0){
		printf("Sector %lu written does not verify properly.\n",sector_number);
		printf("Sector Expected:\n");
		ata_print_sector(new_sector[i]);
		printf("Sector we got:\n");
		ata_print_sector(buf);
		errors++;
	    }
	}
    }
    exit(0);

    /* Now put the sectors back */
    for(i=0;i<max;i++){
	sector_number = 1<<i;
	ret = ata_write_sector(fd,sector_number,1,orig_sector[i]);
	if(ret != ATA_RAW_OK){
	    printf("error encountered writing sector %lu; error code=%d\n",sector_number,ret);
	    errors ++;
	}
    }
    return (errors==0) ? ATA_RAW_OK : ATA_RAW_ERR;
}



int main( void ){
    u_char buff[SECTOR_SIZE];
    u_long generic = 0;
    u_long sector = 0;
    u_char count = 0x01;
    int fd = open("/dev/sda",O_RDWR);
    if( fd <= 2 ) return(-1);
    
    print_test( ata_read_sector( fd, sector, count, buff ), "READ SECTOR" );
    print_test( ata_identify_device( fd, buff ), "IDENTIFY DEVICE" );
    print_test( ata_device_configuration_identify( fd, buff ), "DEVICE CONFIGURATION IDENTIFY" );
    print_test( ata_seek( fd, 500 ), "SEEK" );
    print_test( ata_read_native_max_address( fd, &generic ), "READ NATIVE MAX ADDRESS" );
    print_test (read_write_sector_test(fd), "READ/WRITE SECTOR TEST");


    printf("Successful tests: %d\n",success);
    printf("Failed tests: %d\n",fails);

    /* if failed==0, then we exit OK, otherwise with # of failed tests */
    exit(fails);		
}

